/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 2012, 2018  All Rights Reserved
 */
package java.lang.invoke;

import java.lang.invoke.MethodHandles.Lookup;
import java.lang.reflect.Modifier;

class MethodHandleInfo {

	/* Reference kinds */
	public static final int REF_getField = 1;
	public static final int REF_getStatic = 2;
	public static final int REF_putField = 3;
	public static final int REF_putStatic = 4;
	public static final int REF_invokeVirtual = 5;
	public static final int REF_invokeStatic = 6;
	public static final int REF_invokeSpecial = 7;
	public static final int REF_newInvokeSpecial = 8;
	public static final int REF_invokeInterface = 9;

	private MethodHandleInfoImpl impl;

	/*
	 * Returns information about the specified method handle
	 * through a MethodHandleInfo interface or null if introspection
	 * is not supported on the specified methodHandle
	 * @param methodHandle  The MethodHandle to get information about
	 * @throws Error if MH is non-public method or field in java.lang.invoke
	 */
	public MethodHandleInfo(MethodHandle methodHandle) {
		Class definingClass = methodHandle.definingClass;
		boolean isValid = true;
		String packageName = Lookup.getVMLangAccess().getPackageName(definingClass);
		if ("java.lang.invoke".equals(packageName)) { //$NON-NLS-1$
			/* This MethodHandle is for a field or a method that is in
			 * java.lang.invoke, filter it out if it is not public
			 */
			if (!Modifier.isPublic(methodHandle.rawModifiers)) {
				isValid = false;
			}
		}
		impl = methodHandle.getInfo();
		if ((impl == null) || !isValid) {
			throw new IllegalArgumentException("Invalid MethodHandle for introspection");
		}

	}

	/**
	 * @return class that declares the method or field that the MH points to
	 */
	public Class getDeclaringClass() {
		return impl.getDeclaringClass();
	}

	/**
	 * @return A string representing the name of the method or field
	 */
	public String getName() {
		return impl.getName();
	}

	/**
	 * @return A MethodType object representing the signature of the method or field
	 */
	public MethodType getMethodType() {
		return impl.getMethodType();
	}

	/**
	 * @return An int representing the method or field's modifiers
	 */
	public int getModifiers() {
		return impl.getModifiers();
	}

	/**
	 * @return Returns one of the above defined reference kinds which represent the kind of method handle
	 */
	public int getReferenceKind() {
		return impl.getReferenceKind();
	}

	public String toString() {
		return getReferenceKindString(getReferenceKind()) + " " + getDeclaringClass() + ":" + getMethodType(); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Return a string representing the equivalent bytecode for the referenceKind.
	 * @param referenceKind The referenceKind to lookup
	 * @return a String representing the equivalent bytecode
	 */
	public static String getReferenceKindString(int referenceKind) {
		switch(referenceKind) {
		case REF_getField: 			return "getfield"; //$NON-NLS-1$
		case REF_getStatic: 		return "getstatic"; //$NON-NLS-1$
		case REF_putField: 			return "putfield"; //$NON-NLS-1$
		case REF_putStatic: 		return "putstatic"; //$NON-NLS-1$
		case REF_invokeVirtual: 	return "invokevirtual"; //$NON-NLS-1$
		case REF_invokeStatic: 		return "invokestatic"; //$NON-NLS-1$
		case REF_invokeSpecial: 	return "invokespecial"; //$NON-NLS-1$
		case REF_newInvokeSpecial: 	return "newinvokespecial"; //$NON-NLS-1$
		case REF_invokeInterface: 	return "invokeinterface"; //$NON-NLS-1$
		}
		return "UNKNOWN_REFERENCE_KIND[" + referenceKind + "]"; //$NON-NLS-1$ //$NON-NLS-2$
	}

}

